/*
        Dynamic Stick Gate      ver.1.0 by Radek Kowalski
        
        A better handling for GP2x stock stick using events
        gating in a time domain. (not unlike a noise gate)

        It's a free code and use at your will. I only demand
        to be in app/game credits and nothing more.
        

        There is just one function: DynStickGate what is a basically
        a wrapper for GPIO input. It shoud be called per every
        polling cycle. It has four arguments for its options:
        ord, diag, dyn and prefilter.
        And of course the fifth (or I should say first) argument "in"
        what is a raw input obtained from reading the GPIO.

unsigned int DynStickGate(unsigned int in, unsigned int ord, unsigned int diag,
                                unsigned int dyn, unsigned int prefilter)

        This function returns a processed (gated among other things)
        GPIO inputs. The rest GPIO data flags what aren't necessary
        to change by it are passed.

        Typical implementation of code might look like:

        int GPIO_FileHandler;
        GPIO_FileHandler = open("/dev/GPIO", O_RDWR | O_NDELAY );

        unsigned int Read_Input(void)  // To handle reading from GPIO
        {
            unsigned int buff = 0;

            if(read(GPIO_FileHandler, &buff, 4) != 4)
            {
                  return 0xFFFFFFFF;
            }
            return buff;
        }

        Then main loop of game/app:

        while(!((in = Read_Input()) & GP2X_VK_SELECT)) // exit on Select button
        {
            // whatever
        }

        Adding DynStickGate function is trivial:

        while(!((in = DynStickGate(Read_Input(), 0, 3, 10, 2)) & GP2X_VK_SELECT))
        {
            // whatever
        }

        And it's all to make it works. Of course as it has few parameters
        it'd be preferable to allow to change them in a menu and such.
        But it's a different matter now into its functionality...

        Normally state of GP2x buttons (on/off, 0/1, etc.) is parsed
        immediately when polling input. There isn't a problem with that
        at all. Except the stock stick is somewhat unreliable and its
        bad traits are:
        1. Uneven switch zones resulting in heavy diagonal bias.
            (fixed in "MKII" models of GP2x by rotating stick by 45 degress)
        2. Stock's bar doesn't have any movement constrains resulting in
           analog feel but with discrete output.

        The issue nr. 1 can be deal with DaveC biasing schemes. The best is
        "2" but its effectiveness is reduced by a fact that intermediate
        zones are very small.
        
        The issue nr. 2 (as I know) was never deal with. In a sense it's
        worse than diagonal bias because lack of any constrain results
        in accidental registering of directions. Consider doing UP and
        then wanting to go RIGHT. It's almost impossible to not register
        combi of UP-RIGHT in a way. Moving by 180 degress also
        can suffer from this problem. How it's bad it will depends on
        a game. By a principle the ones what are using upper diagonals
        for jumping will be affected the most. The best example is
        Prince Of Persia running in Outcast Atari ST emulator. Trying
        to play it using stock stick is futile at best
        if not selfmutilating.

        Here comes my idea to do debouncing of stick events. By making
        necessary for diagonals to be in on state longer to register
        accidental states can be filtered out. But it has an expense as
        it will introduce an input latency.
        So it's a compromise however I managed to make it a bit less
        of compromise. Debouncing doesn't have to be on all the time.
        Consider the stick is in neutral postition. It's clear that
        from neutral to any direction the efect of up to right showing
        up, up-right, right is non existant because there no up to right
        in a first place. Therefore debouncing (or a gating) is 
        needed when stick in active state. But it must be set to off
        eventually so it brings a question when? Simple is the best
        - it will shut down after some idle time.

        So back to the DynStickGate function:

unsigned int DynStickGate(unsigned int in, unsigned int ord, unsigned int diag,
                                unsigned int dyn, unsigned int prefilter)
        
        "ord" is a ordinal debouncing (gating) strength
        "diag" is a diagonal debouncing strength

        Both are measured in polling cycles - 0 is no delay for debouncing,
        1 is one extra polling cycle, 2 is two and so forth.

        "dyn" is a number of polling cycles after the debouncing will
        be off when stick is idling. It's counted in polling cycles
        as before. 0 will result in no debouncing at all, a very big
        value (0xFFFFFFFF ;) will result in constant debouncing.

        The "prefilter" is a surprise! To make best of what is possible
        I implemented some DaveC schemes as well. They are done before
        events go into debouncing. There are two of them as now:
        "Case 2" and "Case 4". To use them just pass number 2 or 4 to
        prefilter argument.
        However there is more of prefilter options:
        "7" - is remapping function - the A/B/X/Y will became something
        like a dpad and stick will function as those buttons instead.
        Diagonals are computed by combined buttons push (Y+B will register
        up-right diagonal).
        "8" - the side buttons will act as upper stick diagonals and those
        on the stick (up-left and up-right) will be blocked. Side buttons
        will have a priority over rest of stick directions so going left
        and pushing right sidebutton will result in up-right.
        "9" - it's for stick 4-switch based experiments. It will compute
        diagonals for them so some mods will get a proper functionality.

        For the best DynStickGate results its parameters can be tweaked.
        They will also depends on polling frequency (if it's twice then
        values should also twice to retain the same effect).
        For me (at around 30Hz polling/fps) 0, 2, 5, 2 worked the best.

        The actual code isn't commented. I hadn't time for it, sorry...
*/


// Bit Mask for buttons
#define GP2X_VK_UP              (1 <<  0)
#define GP2X_VK_UP_LEFT         (1 <<  1)
#define GP2X_VK_LEFT            (1 <<  2)
#define GP2X_VK_DOWN_LEFT       (1 <<  3)
#define GP2X_VK_DOWN            (1 <<  4)
#define GP2X_VK_DOWN_RIGHT      (1 <<  5)
#define GP2X_VK_RIGHT           (1 <<  6)
#define GP2X_VK_UP_RIGHT        (1 <<  7)
#define GP2X_VK_START           (1 <<  8)
#define GP2X_VK_SELECT          (1 <<  9)
#define GP2X_VK_FL              (1 << 10)
#define GP2X_VK_FR              (1 << 11)
#define GP2X_VK_FA              (1 << 12)
#define GP2X_VK_FB              (1 << 13)
#define GP2X_VK_FX              (1 << 14)
#define GP2X_VK_FY              (1 << 15)
#define GP2X_VK_VOL_UP          (1 << 16)
#define GP2X_VK_VOL_DOWN        (1 << 17)
#define GP2X_VK_TAT             (1 << 18)

#define GP2X_VK_ALL               0x7FFFF

unsigned int DynStickGate(unsigned int in, unsigned int ord, unsigned int diag,
                                unsigned int dyn, unsigned int prefilter)
{
    static unsigned int UP = 0;
    static unsigned int DOWN = 0;
    static unsigned int LEFT = 0;
    static unsigned int RIGHT = 0;
    static unsigned int UP_LEFT = 0;
    static unsigned int UP_RIGHT = 0;
    static unsigned int DOWN_LEFT = 0;
    static unsigned int DOWN_RIGHT = 0;
    static unsigned int dynamic_gate = 0;
    unsigned int temp;

    ord++;
    diag++;

    if(prefilter == 4)
    {
        if(((in & GP2X_VK_UP) && (in & GP2X_VK_UP_RIGHT)) ||
           ((in & GP2X_VK_UP_RIGHT) && (in & GP2X_VK_RIGHT)) ||
           ((in & GP2X_VK_RIGHT) && (in & GP2X_VK_DOWN_RIGHT)) ||
           ((in & GP2X_VK_DOWN_RIGHT) && (in & GP2X_VK_DOWN)) ||
           ((in & GP2X_VK_DOWN) && (in & GP2X_VK_DOWN_LEFT)) ||
           ((in & GP2X_VK_DOWN_LEFT) && (in & GP2X_VK_LEFT)) ||
           ((in & GP2X_VK_LEFT) && (in & GP2X_VK_UP_LEFT)) ||
           ((in & GP2X_VK_UP_LEFT) && (in & GP2X_VK_UP)))
            in &= ~(GP2X_VK_UP | GP2X_VK_DOWN | GP2X_VK_LEFT | GP2X_VK_RIGHT |
                GP2X_VK_UP_LEFT  | GP2X_VK_UP_RIGHT | GP2X_VK_DOWN_LEFT |
                GP2X_VK_DOWN_RIGHT);
    }
    else if(prefilter == 2)
    {
        if((in & GP2X_VK_UP) && ((in & GP2X_VK_UP_LEFT) || (in & GP2X_VK_UP_RIGHT)))
            in &= ~(GP2X_VK_UP_LEFT | GP2X_VK_UP_RIGHT);

        if((in & GP2X_VK_RIGHT) && ((in & GP2X_VK_UP_RIGHT) || (in & GP2X_VK_DOWN_RIGHT)))
            in &= ~(GP2X_VK_UP_RIGHT | GP2X_VK_DOWN_RIGHT);

        if((in & GP2X_VK_DOWN) && ((in & GP2X_VK_DOWN_LEFT) || (in & GP2X_VK_DOWN_RIGHT)))
            in &= ~(GP2X_VK_DOWN_LEFT | GP2X_VK_DOWN_RIGHT);

        if((in & GP2X_VK_LEFT) && ((in & GP2X_VK_UP_LEFT) || (in & GP2X_VK_DOWN_LEFT)))
            in &= ~(GP2X_VK_UP_LEFT | GP2X_VK_DOWN_LEFT);
    }
    else if(prefilter == 7)
    {
        temp = 0;
        temp |= in & GP2X_VK_FY ? GP2X_VK_UP : 0;
        temp |= in & GP2X_VK_FX ? GP2X_VK_DOWN : 0;
        temp |= in & GP2X_VK_FA ? GP2X_VK_LEFT : 0;
        temp |= in & GP2X_VK_FB ? GP2X_VK_RIGHT : 0;
        temp |= in & GP2X_VK_UP ? GP2X_VK_FY : 0;
        temp |= in & GP2X_VK_DOWN ? GP2X_VK_FX : 0;
        temp |= in & GP2X_VK_LEFT ? GP2X_VK_FA : 0;
        temp |= in & GP2X_VK_RIGHT ? GP2X_VK_FB : 0;
        if((temp & GP2X_VK_UP) && (temp & GP2X_VK_RIGHT))
        {
            temp &= ~(GP2X_VK_UP | GP2X_VK_RIGHT);
            temp |= GP2X_VK_UP_RIGHT;
        }
        if((temp & GP2X_VK_DOWN) && (temp & GP2X_VK_RIGHT))
        {
            temp &= ~(GP2X_VK_DOWN | GP2X_VK_RIGHT);
            temp |= GP2X_VK_DOWN_RIGHT;
        }
        if((temp & GP2X_VK_DOWN) && (temp & GP2X_VK_LEFT))
        {
            temp &= ~(GP2X_VK_DOWN | GP2X_VK_LEFT);
            temp |= GP2X_VK_DOWN_LEFT;
        }
        if((temp & GP2X_VK_UP) && (temp & GP2X_VK_LEFT))
        {
            temp &= ~(GP2X_VK_UP | GP2X_VK_LEFT);
            temp |= GP2X_VK_UP_LEFT;
        }
        in &= ~(GP2X_VK_UP | GP2X_VK_DOWN | GP2X_VK_LEFT | GP2X_VK_RIGHT |
            GP2X_VK_UP_LEFT  | GP2X_VK_UP_RIGHT | GP2X_VK_DOWN_LEFT |
            GP2X_VK_DOWN_RIGHT | GP2X_VK_FY | GP2X_VK_FX | GP2X_VK_FA | GP2X_VK_FB);
        in |= temp;
    }
    else if(prefilter = 8)
    {
        temp = 0;
        temp |= in & GP2X_VK_FL ? GP2X_VK_UP_LEFT : 0;
        temp |= in & GP2X_VK_FR ? GP2X_VK_UP_RIGHT : 0;
        in &= ~(GP2X_VK_FL | GP2X_VK_FR | GP2X_VK_UP_LEFT | GP2X_VK_UP_RIGHT);
        in |= temp;
        if(in & (GP2X_VK_UP_LEFT | GP2X_VK_UP_RIGHT))  
        {
            in &= ~(GP2X_VK_UP | GP2X_VK_DOWN | GP2X_VK_LEFT | GP2X_VK_RIGHT |
                GP2X_VK_DOWN_LEFT | GP2X_VK_DOWN_RIGHT);
            return in;
        }
        if((in & GP2X_VK_RIGHT) && ((in & GP2X_VK_UP_RIGHT) || (in & GP2X_VK_DOWN_RIGHT)))
            in &= ~(GP2X_VK_UP_RIGHT | GP2X_VK_DOWN_RIGHT);

        if((in & GP2X_VK_DOWN) && ((in & GP2X_VK_DOWN_LEFT) || (in & GP2X_VK_DOWN_RIGHT)))
            in &= ~(GP2X_VK_DOWN_LEFT | GP2X_VK_DOWN_RIGHT);

        if((in & GP2X_VK_LEFT) && ((in & GP2X_VK_UP_LEFT) || (in & GP2X_VK_DOWN_LEFT)))
            in &= ~(GP2X_VK_UP_LEFT | GP2X_VK_DOWN_LEFT);
    }
    else if(prefilter = 9)
    {
        if((in & GP2X_VK_UP) && (in & GP2X_VK_RIGHT))
        {
            in &= ~(GP2X_VK_UP | GP2X_VK_RIGHT);
            in |= GP2X_VK_UP_RIGHT;
        }
        if((in & GP2X_VK_DOWN) && (in & GP2X_VK_RIGHT))
        {
            in &= ~(GP2X_VK_DOWN | GP2X_VK_RIGHT);
            in |= GP2X_VK_DOWN_RIGHT;
        }
        if((in & GP2X_VK_DOWN) && (in & GP2X_VK_LEFT))
        {
            in &= ~(GP2X_VK_DOWN | GP2X_VK_LEFT);
            in |= GP2X_VK_DOWN_LEFT;
        }
        if((in & GP2X_VK_UP) && (in & GP2X_VK_LEFT))
        {
            in &= ~(GP2X_VK_UP | GP2X_VK_LEFT);
            in |= GP2X_VK_UP_LEFT;
        }
    }

    if(prefilter > 0)
    {
        temp = 0;
        temp += in & GP2X_VK_UP ? 1 : 0;
        temp += in & GP2X_VK_UP_RIGHT ? 1 : 0;
        temp += in & GP2X_VK_RIGHT ? 1 : 0;
        temp += in & GP2X_VK_DOWN_RIGHT ? 1 : 0;
        temp += in & GP2X_VK_DOWN ? 1 : 0;
        temp += in & GP2X_VK_DOWN_LEFT ? 1 : 0;
        temp += in & GP2X_VK_LEFT ? 1 : 0;
        temp += in & GP2X_VK_UP_LEFT ? 1 : 0;
        if(temp > 1)
            in &= ~(GP2X_VK_UP | GP2X_VK_DOWN | GP2X_VK_LEFT | GP2X_VK_RIGHT |
                GP2X_VK_UP_LEFT  | GP2X_VK_UP_RIGHT | GP2X_VK_DOWN_LEFT |
                GP2X_VK_DOWN_RIGHT);
    }

    if(dynamic_gate == 0)
    {
        ord = 1;
        diag = 1;
    }

    UP = in & GP2X_VK_UP ? (UP > 0 ? UP - 1 : 0) : ord;
    DOWN = in & GP2X_VK_DOWN ? (DOWN > 0 ? DOWN - 1 : 0) : ord;
    LEFT = in & GP2X_VK_LEFT ? (LEFT > 0 ? LEFT - 1 : 0) : ord;
    RIGHT = in & GP2X_VK_RIGHT ? (RIGHT > 0 ? RIGHT - 1 : 0) : ord;
    UP_RIGHT = in & GP2X_VK_UP_RIGHT ? (UP_RIGHT > 0 ? UP_RIGHT - 1 : 0) : diag;
    UP_LEFT = in & GP2X_VK_UP_LEFT ? (UP_LEFT > 0 ? UP_LEFT - 1 : 0) : diag;
    DOWN_LEFT = in & GP2X_VK_DOWN_LEFT ? (DOWN_LEFT > 0 ? DOWN_LEFT - 1 : 0) : diag;
    DOWN_RIGHT = in & GP2X_VK_DOWN_RIGHT ? (DOWN_RIGHT > 0 ? DOWN_RIGHT - 1 : 0) : diag;

    dynamic_gate = in & (GP2X_VK_UP | GP2X_VK_DOWN | GP2X_VK_LEFT | GP2X_VK_RIGHT |
        GP2X_VK_UP_LEFT  | GP2X_VK_UP_RIGHT | GP2X_VK_DOWN_LEFT |
        GP2X_VK_DOWN_RIGHT) ? dyn : (dynamic_gate > 0 ? dynamic_gate - 1 : 0);

    in &= ~(GP2X_VK_UP | GP2X_VK_DOWN | GP2X_VK_LEFT | GP2X_VK_RIGHT |
        GP2X_VK_UP_LEFT  | GP2X_VK_UP_RIGHT | GP2X_VK_DOWN_LEFT |
        GP2X_VK_DOWN_RIGHT);
    in |= UP ? 0 : GP2X_VK_UP;
    in |= DOWN ? 0 : GP2X_VK_DOWN;
    in |= LEFT ? 0 : GP2X_VK_LEFT;
    in |= RIGHT ? 0 : GP2X_VK_RIGHT;
    in |= UP_LEFT ? 0 : GP2X_VK_UP_LEFT;
    in |= UP_RIGHT ? 0 : GP2X_VK_UP_RIGHT;
    in |= DOWN_LEFT ? 0: GP2X_VK_DOWN_LEFT;
    in |= DOWN_RIGHT ? 0: GP2X_VK_DOWN_RIGHT;

    return in;
}
